using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;

namespace System.IO
{
    public class DataStream
    {
        public Stream BaseStream { get { return _baseStream; } }
        protected Stream _baseStream;
        protected BinaryReader br;
        protected BinaryWriter bw;

        public DataStream(Stream stream)
        {
            _baseStream = stream;
            br = new BinaryReader(_baseStream);
            bw = new BinaryWriter(_baseStream);
            _bookmarks = new Stack<long>();
        }

        #region Bookmarks
        public void PushBookmark()
        { _bookmarks.Push(_baseStream.Position); }

        public void PopBookmark()
        { _baseStream.Position = _bookmarks.Pop(); }
        private Stack<long> _bookmarks;
        #endregion

        #region Read

        public T Read<T>() where T : IDatatizable
        {
            return (T)typeof(T).InvokeMember(null,
         BindingFlags.DeclaredOnly |
         BindingFlags.Public | BindingFlags.NonPublic |
         BindingFlags.Instance | BindingFlags.CreateInstance, null, null, new object[] { this });
        }

        public bool ReadBool()
        { return br.ReadBoolean(); }

        public byte ReadByte()
        { return br.ReadByte(); }

        public byte[] ReadBytes(int count)
        { return br.ReadBytes(count); }

        public MemoryStream ReadMemoryStream(int byteCount)
        { return new MemoryStream(br.ReadBytes(byteCount)); }

        public char ReadChar()
        { return br.ReadChar(); }

        public char[] ReadChars(int count)
        { return br.ReadChars(count); }

        public decimal ReadDecimal()
        { return br.ReadDecimal(); }

        public double ReadDouble()
        { return br.ReadDouble(); }

        public short ReadInt16()
        { return br.ReadInt16(); }

        public int ReadInt32()
        { return br.ReadInt32(); }

        public long ReadInt64()
        { return br.ReadInt64(); }

        public sbyte ReadSByte()
        { return br.ReadSByte(); }

        public float ReadSingle()
        { return br.ReadSingle(); }

        #region public string ReadString()

        public string ReadStringCharacters(int count)
        {
            return ReadStringCharacters(count, false);
        }
        public string ReadStringCharacters(int count, bool reverse)
        { return (reverse) ? ReverseString(new string(br.ReadChars(count))) : new string(br.ReadChars(count)); }

        public string ReadStringNullTerminated()
        { return ReadStringNullTerminated(false); }
        public string ReadStringNullTerminated(bool reverse)
        {
            char c;
            string value = "";
            do
            {
                c = br.ReadChar();
                value += c;
            } while (c != '\0');
            return (reverse) ? ReverseString(value) : value;
        }

        public string ReadStringSystemIO()
        { return ReadStringSystemIO(false); }
        public string ReadStringSystemIO(bool reverse)
        { return (reverse) ? ReverseString(br.ReadString()) : br.ReadString(); }

        private string ReverseString(string value)
        {
            char[] chars = value.ToCharArray();
            Array.Reverse(chars);
            return new string(chars);
        }

        #endregion

        public ushort ReadUInt16()
        { return br.ReadUInt16(); }

        public uint ReadUInt32()
        { return br.ReadUInt32(); }

        public ulong ReadUInt64()
        { return br.ReadUInt64(); }

        public Flags ReadFlags8()
        { return new Flags8(this); }

        public Flags ReadFlags16()
        { return new Flags16(this); }

        public Flags ReadFlags32()
        { return new Flags32(this); }

        public Flags ReadFlags64()
        { return new Flags64(this); }

        #endregion
        #region Write

        public void Write(IDatatizable value)
        { value.Write(this); }

        public void Write(bool value)
        { bw.Write(value); }

        public void Write(byte value)
        { bw.Write(value); }

        public void Write(byte[] buffer)
        { bw.Write(buffer); }

        public void Write(Stream buffer)
        { 
            byte[] bytes = new byte[buffer.Length];
            buffer.Read(bytes, 0, bytes.Length);
            bw.Write(bytes);
        }

        public void Write(char ch)
        { bw.Write(ch); }

        public void Write(char[] chars)
        { bw.Write(chars); }

        public void Write(decimal value)
        { bw.Write(value); }

        public void Write(double value)
        { bw.Write(value); }

        public void Write(float value)
        { bw.Write(value); }

        public void Write(int value)
        { bw.Write(value); }

        public void Write(long value)
        { bw.Write(value); }

        public void Write(sbyte value)
        { bw.Write(value); }

        public void Write(short value)
        { bw.Write(value); }

        #region public void Write(string value)

        public void WriteStringCharacters(string value)
        { WriteStringCharacters(value, false); }
        public void WriteStringCharacters(string value, bool reverse)
        { bw.Write((reverse) ? ReverseString(value).ToCharArray() : value.ToCharArray()); }

        public void WriteStringNullTerminated(string value)
        { WriteStringNullTerminated(value, false); }
        public void WriteStringNullTerminated(string value, bool reverse)
        {
            bw.Write((reverse) ? ReverseString(value).ToCharArray() : value.ToCharArray());
            bw.Write('\0');
        }

        public void WriteStringSystemIO(string value)
        { WriteStringSystemIO(value, false); }
        public void WriteStringSystemIO(string value, bool reverse)
        { bw.Write((reverse) ? ReverseString(value) : value); }

        #endregion

        public void Write(uint value)
        { bw.Write(value); }

        public void Write(ulong value)
        { bw.Write(value); }

        public void Write(ushort value)
        { bw.Write(value); }

        public void Write(byte[] buffer, int index, int count)
        { bw.Write(buffer, index, count); }

        public void Write(char[] chars, int index, int count)
        { bw.Write(chars, index, count); }

        public void Write(Flags value)
        { value.Write(this); }

        #endregion
    }
}
